/*
 * Mailfront Plugin Name: plugin-check-helo
 *
 * Reference:
 * http://www.linuxmagic.com/best_practices/valid_helo_domain.html
 *
 * This plugin is used to check HELO name and remote hostname.
 *
 * This plugin checks sequence:
 *
 * (1).  If $HELO_CHECK_SKIP is set, this plugin will be skipped.
 *
 * (2).  If $RELAYCLIENT is set, add to message header and skip the remaining
 *       checks.  Example of message header added:
 *
 *       X-HELO-Check-Result: SKIPPED: $RELAYCLIENT is set for 1.2.3.4
 *
 * (3).  If sender is authenticated, add to message header and skip the
 *       remaining checks.  Example of message header added:
 *
 *       X-HELO-Check-Result: SKIPPED: authenticated user from 1.2.3.4
 *
 * (4).  If $RBLSMTPD is set and empty, add to message header and skip the
 *       remaining checks.
 *
 * (5).  If $HELO_CHECK_REMOTEHOST or $HELO_CHECK_REMOTEHOST_MATCH is set,
 *       this plugin will check ${PROTO}REMOTEHOST is not empty.  If is empty,
 *       will do a DNS PTR lookup using ${PROTO}REMOTEIP and use the return
 *       value as ${PROTO}REMOTEHOST.
 *
 * (6).  If $HELO_CHECK_REMOTEHOST is set, this plugin will check HELO name
 *       is not empty.  If it is empty, add to message header or defer or
 *       reject accordingly and skip the remaining checks.
 *       (Refer to $HELO_CHECK_ADD_HEADER, $HELO_CHECK_DEFER and
 *       $HELO_CHECK_REJECT)
 *       Example of message header added:
 *
 *       X-HELO-Check-Result: FAILED
 *       X-HELO-Check-Summary: REMOTEHOST is empty and PTR DNS lookup for IP 1.2.3.4 failed
 *
 * (7).  If $HELO_CHECK_REMOTEHOST_MATCH is set, this plugin will check HELO
 *       name and ${PROTO}REMOTEHOST whether both are the same otherwise add
 *       to message header or defer or reject accordingly and skip the
 *       remaining checks.
 *       (Refer to $HELO_CHECK_ADD_HEADER, $HELO_CHECK_DEFER and
 *       $HELO_CHECK_REJECT)
 *       Example of message header added:
 *
 *       X-HELO-Check-Result: FAILED
 *       X-HELO-Check-Summary: HELO name and REMOTEHOST doesn't match
 *
 * (8).  Check HELO name for the following:
 *       (8.1).  Cannot starts with ".", "[" and "-"
 *       (8.2).  Cannot ends with ".", "]" and "-"
 *       (8.3).  Must have at least a "." (dot)
 *       (8.4).  Cannot have any of the following characters besides TAB
 *               and SPACE:
 *                   ~!@#$%^&*()_+={}[]:;"'<>,?/|
 *       (8.5).  Cannot use a valid IPv4 address
 *       NOTE: If any of the above checks failed, the remaining checks will
 *       be skipped.
 *       (Refer to $HELO_CHECK_ADD_HEADER, $HELO_CHECK_DEFER and
 *       $HELO_CHECK_REJECT)
 *       Example of message header added:
 *
 *       X-HELO-Check-Result: FAILED
 *       X-HELO-Check-Summary: '.' found in last character of HELO name sender_set_HELO_name
 *
 * (9).  Do a DNS lookup for HELO name for A, AAAA and PTR records.  If the
 *       DNS lookup failed due to no such name exists will skip all remaining
 *       checks and add to message header or defer or reject.
 *       (Refer to $HELO_CHECK_ADD_HEADER, $HELO_CHECK_DEFER and
 *       $HELO_CHECK_REJECT)
 *       Any other errors such as timeout etc. will result in skipping
 *       the remaining checks and message will be deferred with reply code
 *       451 if $HELO_CHECK_REJECT or $HELO_CHECK_DEFER is set otherwise
 *       if $HELO_CHECK_ADD_HEADER is set will add to message header.
 *       Example of message header added:
 *
 *       X-HELO-Check-Result: FAILED
 *       X-HELO-Check-Summary: DNS lookup error for HELO name sender_set_HELO_name
 *
 * (10). If $HELO_CHECK_REMOTEIP is set, this plugin will check for a match
 *       of ${PROTO}REMOTEIP and HELO name IP.  If there is a match, skip
 *       the remaining checks and add header if $HELO_CHECK_ADD_HEADER is set.
 *       If there is no match, add to message header or defer or reject.
 *       (Refer to $HELO_CHECK_ADD_HEADER, $HELO_CHECK_DEFER and
 *       $HELO_CHECK_REJECT)
 *       Example of message header added:
 *
 *       X-HELO-Check-Result: PASSED
 *       X-HELO-Check-Summary: REMOTEIP matches sender_set_HELO_name's IP
 *
 * (11). If $HELO_CHECK_REMOTEIP_MATCH is set, this plugin will check for a
 *       match the first 3 octets (within same class C) of ${PROTO}REMOTEIP
 *       if is a valid IPv4 address against helo name A records.  If there
 *       isn't a match, add headers or defer or reject accordingly.
 *       If there is a match, add headers if $HELO_CHECK_ADD_HEADER and skip
 *       the rest of the checks.
 *       Example of message header added:
 *
 *       X-HELO-Check-Result: SKIPPED
 *       X-HELO-Check-Summary: REMOTEIP 1.2.3.4 matches first 3 octets of sender_set_HELO_name's IP 1.2.3.50 (same class C)
 *
 * (12). If $HELO_CHECK_PTR is set, this plugin will check the HELO name's
 *       A records to have at least one PTR record. If the check failed,
 *       it will add to message header or defer or reject accordingly.
 *       Example of message header added:
 *
 *       X-HELO-Check-Result: FAILED
 *       X-HELO-Check-Summary: X-HELO-Check-Summary: PTR record not found for HELO name sender_set_HELO_name
 *
 * (13). If $HELO_CHECK_PTR_MATCH is set, this plugin will check there is at
 *       least one of the PTR record resolves to the same given HELO name.
 *       If the check failed, it will add to message header or defer or reject accordingly.
 *       Example of message header added:
 *
 *       X-HELO-Check-Result: FAILED
 *       X-HELO-Check-Summary: X-HELO-Check-Summary: PTR record doesn't resolves back to HELO name sender_set_HELO_name
 *
 * (14). If $HELO_CHECK_REJECT is set, sender will be rejected with reply
 *       code 553 if any of the above checks failed.
 *
 * (15). If $HELO_CHECK_DEFER is set, sender will be deferred with reply
 *       code 451 if any of the above checks failed.
 *
 * (16). If $HELO_CHECK_ADD_HEADER is set, this plugin will add its checking
 *       result to each mail message header accepted.
 *
 * (17). If $HELO_CHECK_VERBOSE is set, this plugin will log all its checking
 *       to stderr.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <resolv.h>

#include <msg/msg.h>
#include "mailfront.h"

#define eprintf(...) fprintf(stderr, __VA_ARGS__)
#define elogerr(N, M, ...) { if(N){ if(M){ eprintf_logname(); } fprintf(stderr, __VA_ARGS__); } }

#ifndef MAX_DNSIP_RECORDS
#define MAX_DNSIP_RECORDS 10
#endif

static response resp;

static const char *mainprog = "mailfront";
static const char *thisprog = "plugin-check-helo";
static unsigned long thispid;
static const char *local_host;
static const char *local_ip;
static const char *remote_host;
static const char *remote_ip;
static const char *relayclient;
static const char *rejectcheck;
static const char *defercheck;
static const char *skipcheck;
static const char *logcheck;
static const char *addheadercheck;
static const char *checkremotehost;
static const char *checkremotehostmatch;
static const char *checkremoteip;
static const char *checkremoteipmatch;
static int outofmemory = 0;
static str estr;
static str hmsg;

void eprintf_logname()
{
  if (!thispid) {
    thispid = (unsigned long)getpid();
  }
  eprintf("%s-%s[%lu]: ", mainprog, thisprog, thispid);
}

struct HostDNSIP {
  char *hostname;
  int size;
  int ptrcheck;
  int ipv4count;
  int ipv6count;
  int ipcount;
  char *ip[MAX_DNSIP_RECORDS];
  int ptrcount;
  char *ptr[MAX_DNSIP_RECORDS];
};

void init_HostDNSIP(struct HostDNSIP *d)
{
  d->hostname = NULL;
  d->size = 0;
  d->ptrcheck = 0;
  d->ipv4count = 0;
  d->ipv6count = 0;
  d->ipcount = 0;
  d->ptrcount = 0;
}

void free_HostDNSIP(struct HostDNSIP *d)
{
  int x = 0;
  if (d->ipcount > 0) {
    for (x = 0; x < d->ipcount; x++) {
      free(d->ip[x]);
    }
    d->ipcount = 0;
  }
  if (d->ptrcount > 0) {
    for (x = 0; x < d->ptrcount; x++) {
      free(d->ptr[x]);
    }
    d->ptrcount = 0;
  }
  init_HostDNSIP(d);
}

/* return NULL if out of memory */
char *ip_ptr_lookup(char *ip)
{
  struct hostent *hent;
  struct in_addr addr;

  char *hostname = (char *)malloc(256);
  if (hostname == NULL) {
    return NULL;
  }
  sprintf(hostname, "%s", ip);

  if(!inet_aton(ip, &addr))
    return hostname;

  if(strchr(ip, ':')) {
    if((hent = gethostbyaddr((char *)&(addr.s_addr), sizeof(addr.s_addr), AF_INET6))) {
      strcpy(hostname, hent->h_name);
    }
  } else {
    if((hent = gethostbyaddr((char *)&(addr.s_addr), sizeof(addr.s_addr), AF_INET))) {
      strcpy(hostname, hent->h_name);
    }
  }
  return hostname;
}

int is_valid_ipv4(const char *s)
{
  unsigned int n1, n2, n3, n4;

  if (sscanf(s, "%3u.%3u.%3u.%3u", &n1, &n2, &n3, &n4) != 4) return 0;

  if ((n1 != 0) && (n1 <= 255) && (n2 <= 255) && (n3 <= 255) && (n4 <= 255)) {
    char buf[64];
    sprintf(buf, "%u.%u.%u.%u", n1, n2, n3, n4);
    if(strcmp(buf, s)) return 0;
    return 1;
  }
  return 0;
}

/* Match number of octets in ipv4 address and return 1 if match
 * If IP is within the same class A/B/C return 1 depending on octets value
 */
int match_ipv4(const char *ip1, const char *ip2, int octets)
{
  unsigned int a1, a2, a3, a4;
  unsigned int b1, b2, b3, b4;

  if (!is_valid_ipv4(ip1)) return 0;
  if (!is_valid_ipv4(ip2)) return 0;

  if (sscanf(ip1, "%3u.%3u.%3u.%3u", &a1, &a2, &a3, &a4) != 4) return 0;
  if (sscanf(ip2, "%3u.%3u.%3u.%3u", &b1, &b2, &b3, &b4) != 4) return 0;
  if (octets == 4) {
    if (a1 == b1 && a2 == b2 && a3 == b3 && a4 == b4) return 1;
  }
  if (octets == 3) {
    if (a1 == b1 && a2 == b2 && a3 == b3) return 1;
  }
  if (octets == 2) {
    if (a1 == b1 && a2 == b2) return 1;
  }
  if (octets == 1) {
    if (a1 == b1) return 1;
  }
  return 0;
}

void str_tolower(const char *a, char *b) {
  int i, len = strlen(a);
  for(i = 0; i < len; i++) {
    b[i] = (char)tolower(a[i]);
  }
  b[len] = '\0';
}

/*
 * return 1 if hostname DNS lookup has A record IPv4 or IPv6
 * otherwise return 0 if failed or out of memory which will increment
 * static int outofmemory
 */
int check_hostname(const char *hostname, struct HostDNSIP *s, int limit)
{
  struct addrinfo hints, *res, *result;
  int ret = 0;
  int found = 0;
  int i = 0;
  int c = 0;
  if (hostname) {
    if (s->hostname) {
      elogerr(logcheck, 1, "(%s:%d) s->hostname already set to %s\n", __FILE__, __LINE__, s->hostname);
      return 1;
    }
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC; // Handle IPv4 and IPv6
    hints.ai_socktype = SOCK_STREAM;
    ret = getaddrinfo(hostname, NULL, &hints, &result);
    if(ret) {
      elogerr(logcheck, 1, "(%s:%d) hostname=%s errno=%d errmsg=%s\n", __FILE__, __LINE__, hostname, ret, gai_strerror(ret));
      return ret;
    } else {
      for (c = 0, res = result; res != NULL && c < limit; res = res->ai_next, c++) {
        char ipstr[64];
        int x = 0;
        int saved = 1;
        struct sockaddr_in *ipv = (struct sockaddr_in *)res->ai_addr;
        void * addr = &(ipv->sin_addr);
        inet_ntop(res->ai_family, addr, ipstr, sizeof(ipstr));
        if (strcmp(ipstr, hostname)) {
          s->hostname = (char *)hostname;
          found = 1;
        }
        if (i > 0) {
          for (x = 0; x < i; x++) {
            if (strcmp(s->ip[x],
              ipstr) == 0) {
              saved = 0;
            }
          }
        }
        if (saved > 0) {
          s->ip[i] = strdup(ipstr);
          if (s->ip[i] == NULL) {
            elogerr(1, 1, "(%s:%d) Out of memory!\n", __FILE__, __LINE__);
            outofmemory++;
            return 0;
          }
          s->ipcount += 1;
          s->ptr[i] = ip_ptr_lookup(ipstr);
          if (s->ptr[i] == NULL) {
            elogerr(1, 1, "(%s:%d) Out of memory!\n", __FILE__, __LINE__);
            outofmemory++;
            return 0;
          }
          s->ptrcount += 1;
          /* If ip and ptr are the same means PTR lookup failed
           * If hostname and ptr are the same means ptr resolves back to the same hostname
           */
          if (strcmp(s->ip[i], s->ptr[i])) {
            if (strcmp(s->hostname, s->ptr[i]) == 0) {
              s->ptrcheck++;
            }
          }
          /* See ip is ipv6 by looking for character ':' */
          if (strchr(s->ip[i], ':')) {
            s->ipv6count++;
          } else {
            s->ipv4count++;
          }
          i++;
        }
      }
      s->size = i;
      freeaddrinfo(result);
    }
  }
  if (found) {
    return found;
  }
  return 0;
}

static const response *init(void)
{
  local_ip = getprotoenv("LOCALIP");
  remote_ip = getprotoenv("REMOTEIP");
  local_host = getprotoenv("LOCALHOST");
  remote_host = getprotoenv("REMOTEHOST");
  relayclient = session_getenv("RELAYCLIENT");
  rejectcheck = session_getenv("HELO_CHECK_REJECT");
  defercheck = session_getenv("HELO_CHECK_DEFER");
  skipcheck = session_getenv("HELO_CHECK_SKIP");
  checkremotehost = session_getenv("HELO_CHECK_REMOTEHOST");
  checkremotehostmatch = session_getenv("HELO_CHECK_REMOTEHOST_MATCH");
  checkremoteip = session_getenv("HELO_CHECK_REMOTEIP");
  checkremoteipmatch = session_getenv("HELO_CHECK_REMOTEIP_MATCH");
  if (session_getenv("SENDER_HELO_HOSTNAME") != 0) {
    unsetenv("SENDER_HELO_HOSTNAME");
  }
  logcheck = session_getenv("HELO_CHECK_VERBOSE");
  addheadercheck = session_getenv("HELO_CHECK_ADD_HEADER");
  thispid = (unsigned long)getpid();
  return 0;
}

static const response *helo(str *hostname, str *capabilities)
{
  if (session_getenv("SENDER_HELO_HOSTNAME") == 0) {
    char t[strlen(hostname->s) + 1];
    str_tolower(hostname->s, t);
    session_setenv("SENDER_HELO_HOSTNAME", t, 0);
  }
  return 0;
  (void)capabilities;
}

int check_senderhelo_char_start(const char *s, int i, char c)
{
  char d = '\'';
  if (c == '\'') {
    d = '"';
  }
  /* first character */
  if (*s == c) {
    if (i > 0) {
      char *tmp = (char *)malloc(100);
      if (tmp == NULL) {
        outofmemory++;
        return 0;
      }
      sprintf(tmp, "Sorry, your HELO name starts with disallowed character %c%c%c", d, c, d);
      str_init(&estr);
      str_cats(&estr, tmp);
      resp.message = estr.s;
      free(tmp);
      return 2;
    } else {
      elogerr(logcheck, 1, "%c%c%c found in first character of HELO name %s\n", d, c, d, s);
      str_cats(&hmsg, "FAILED\n");
      str_cats(&hmsg, "X-HELO-Check-Summary: ");
      str_catc(&hmsg, d);
      str_catc(&hmsg, c);
      str_catc(&hmsg, d);
      str_cats(&hmsg, " found in first character of HELO name ");
      str_cats(&hmsg, s);
      str_cats(&hmsg, "\n");
      return 1;
    }
  }
  return 0;
}

int check_senderhelo_char_end(const char *s, int i, char c)
{
  char d = '\'';
  if (c == '\'') {
    d = '"';
  }
  int len = strlen(s);
  /* last character */
  if (s[len-1] == c) {
    if (i > 0) {
      char *tmp = (char *)malloc(100);
      if (tmp == NULL) {
        outofmemory++;
        return 0;
      }
      sprintf(tmp, "Sorry, your HELO name ends with disallowed character %c%c%c", d, c, d);
      str_init(&estr);
      str_cats(&estr, tmp);
      resp.message = estr.s;
      free(tmp);
      return 2;
    } else {
      elogerr(logcheck, 1, "%c%c%c found in last character of HELO name %s\n", d, c, d, s);
      str_cats(&hmsg, "FAILED\n");
      str_cats(&hmsg, "X-HELO-Check-Summary: ");
      str_catc(&hmsg, d);
      str_catc(&hmsg, c);
      str_catc(&hmsg, d);
      str_cats(&hmsg, " found in last character of HELO name ");
      str_cats(&hmsg, s);
      str_cats(&hmsg, "\n");
      return 1;
    }
  }
  return 0;
}

int check_senderhelo_char(const char *s, int i, char c)
{
  if (strchr(s, c)) {
    char d = '\'';
    if (c == '\'') {
      d = '"';
    }
    if (i > 0) {
      char *tmp = (char *)malloc(100);
      if (tmp == NULL) {
        outofmemory++;
        return 0;
      }
      sprintf(tmp, "Sorry, your HELO name has the disallowed character %c%c%c", d, c, d);
      str_init(&estr);
      str_cats(&estr, tmp);
      resp.message = estr.s;
      free(tmp);
      return 2;
    } else {
      elogerr(logcheck, 1, "%c%c%c character found in HELO name %s\n", d, c, d, s);
      str_cats(&hmsg, "FAILED\n");
      str_cats(&hmsg, "X-HELO-Check-Summary: ");
      str_catc(&hmsg, d);
      str_catc(&hmsg, c);
      str_catc(&hmsg, d);
      str_cats(&hmsg, " character found in HELO name ");
      str_cats(&hmsg, s);
      str_cats(&hmsg, "\n");
      return 1;
    }
  }
  return 0;
}

static const response *sender(str* address, str* params)
{
  if (skipcheck) return 0;
  int action = 0;
  if (rejectcheck || defercheck) {
    if (rejectcheck) {
      action = 553;
      resp.number = action;
    } else {
      action = 451;
      resp.number = action;
    }
  }
  if (!local_ip) {
    local_ip = getprotoenv("LOCALIP");
  }
  if (!local_ip) {
    elogerr(logcheck, 1, "LOCALIP not found thus this plugin skipped\n");
    return 0;
  }
  local_host = getprotoenv("LOCALHOST");
  if (!remote_ip) {
    remote_ip = getprotoenv("REMOTEIP");
  }
  if (!remote_ip) {
    elogerr(logcheck, 1, "REMOTEIP not found thus this plugin skipped\n");
    return 0;
  }
  str_init(&hmsg);
  str_copys(&hmsg, "X-HELO-Check-Result: ");
  /* X-HELO-Check-Result: SKIPPED: $RELAYCLIENT is set for 1.2.3.4 */
  if (relayclient) {
    elogerr(logcheck, 1, "$RELAYCLIENT is set thus skipped for REMOTEIP %s\n", remote_ip);
    str_cats(&hmsg, "SKIPPED: $RELAYCLIENT is set for ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  /* X-HELO-Check-Result: SKIPPED: authenticated user from 1.2.3.4 */
  if (session_getnum("authenticated", 0)) {
    elogerr(logcheck, 1, "authenticated user thus skipped for REMOTEIP %s\n", remote_ip);
    str_cats(&hmsg, "SKIPPED: authenticated user from ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  /* X-HELO-Check-Result: SKIPPED: LOCALIP is the same as REMOTEIP 1.2.3.4 */
  if (strcmp(local_ip, remote_ip) == 0) {
    elogerr(logcheck, 1, "LOCALIP=REMOTEIP=%s thus skipped\n", remote_ip);
    str_cats(&hmsg, "SKIPPED: LOCALIP is the same as REMOTEIP ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  /* X-HELO-Check-Result: SKIPPED: $RBLSMTPD is set and empty for 1.1.1.1 */
  const char *rblsmtpd = getenv("RBLSMTPD");
  if (rblsmtpd && *rblsmtpd == 0) {
    elogerr(logcheck, 1, "$RBLSMTPD is set and empty thus skipped for REMOTEIP %s\n", remote_ip);
    str_cats(&hmsg, "SKIPPED: $RBLSMTPD is set and empty for ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  if (!remote_host) {
    remote_host = getprotoenv("REMOTEHOST");
  }
  if ((checkremotehost || checkremotehostmatch) && (!remote_host || *remote_host == 0)) {
    if (checkremotehost && checkremotehostmatch) {
      elogerr(logcheck, 1, "$HELO_CHECK_REMOTEHOST and $HELO_CHECK_REMOTEHOST_MATCH are set for REMOTEIP %s\n", remote_ip);
    } else if (checkremotehost) {
      elogerr(logcheck, 1, "$HELO_CHECK_REMOTEHOST is set for REMOTEIP %s\n", remote_ip);
    } else {
      elogerr(logcheck, 1, "$HELO_CHECK_REMOTEHOST_MATCH is set for REMOTEIP %s\n", remote_ip);
    }
    elogerr(logcheck, 1, "REMOTEHOST is empty thus trying to do PTR DNS lookup from REMOTEIP %s\n", remote_ip);
    char *tmphost = ip_ptr_lookup((char *)remote_ip);
    if (outofmemory > 0 || tmphost == NULL) {
      elogerr(1, 1, "(%s:%d) char *tmphost = ip_ptr_lookup((char *)remote_ip) failed!\n", __FILE__, __LINE__);
      return &resp_oom;
    }
    if (strcmp(tmphost, remote_ip) != 0 && strlen(tmphost) > 0) {
      remote_host = strdup(tmphost);
      if (remote_host == NULL) {
        elogerr(1, 1, "(%s:%d) remote_host=strdup(tmphost) failed!\n", __FILE__, __LINE__);
        free(tmphost);
        return &resp_oom;
      }
      elogerr(logcheck, 1, "PTR DNS loopkup of REMOTEIP=%s success and set REMOTEHOST=%s\n", remote_ip, tmphost);
      free(tmphost);
    } else {
      free(tmphost);
      elogerr(logcheck, 1, "PTR DNS loopkup for REMOTEIP %s failed\n", remote_ip);
      if (action > 0) {
        str_init(&estr);
        str_cats(&estr, "Sorry, your hostname is empty and your ip ");
        str_cats(&estr, remote_ip);
        str_cats(&estr, " doesn't have any PTR record");
        resp.message = estr.s;
        return &resp;
      } else {
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEHOST is empty and PTR DNS lookup for IP ");
        str_cats(&hmsg, remote_ip);
        str_cats(&hmsg, " failed\n");
        return 0;
      }
    }
  }
  const char *senderhelo = session_getenv("SENDER_HELO_HOSTNAME");
  /* If helo_hostname is empty */
  if (senderhelo == 0 || *senderhelo == 0) {
    if (action > 0) {
      resp.message = "Sorry, your HELO name is empty!";
      return &resp;
    } else {
      elogerr(logcheck, 1, "helo hostname is empty\n");
      str_cats(&hmsg, "FAILED\n");
      str_cats(&hmsg, "X-HELO-Check-Summary: HELO name is empty!\n");
      return 0;
    }
  }
  if (senderhelo && *senderhelo != 0) {
    if (checkremotehostmatch) {
      if (strcmp(senderhelo, remote_host) != 0) {
        elogerr(logcheck, 1, "HELO name %s and REMOTEHOST %s doesn't match for REMOTEIP %s\n", senderhelo, remote_host, remote_ip);
        if (action > 0) {
          str_init(&estr);
          str_cats(&estr, "Sorry, your HELO name ");
          str_cats(&estr, senderhelo);
          str_cats(&estr, " and REMOTEHOST ");
          str_cats(&estr, remote_host);
          str_cats(&estr, " doesn't match");
          resp.message = estr.s;
          return &resp;
        } else {
          str_cats(&hmsg, "FAILED\n");
          str_cats(&hmsg, "X-HELO-Check-Summary: HELO name ");
          str_cats(&hmsg, senderhelo);
          str_cats(&hmsg, " REMOTEHOST ");
          str_cats(&hmsg, remote_host);
          str_cats(&hmsg, " doesn't match\n");
          return 0;
        }
      } else {
        elogerr(logcheck, 1, "HELO name %s and REMOTEHOST %s match for REMOTEIP %s\n", senderhelo, remote_host, remote_ip);
      }
    }
    int len = strlen(senderhelo);
    /* 0 continue, 1 to return 0 and 2 to return &resp; */
    int rc = 0;
    /* Don't accept start with a dot */
    rc = check_senderhelo_char_start(senderhelo, action, '.');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    /* Don't accept start with a [ */
    rc = check_senderhelo_char_start(senderhelo, action, '[');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    /* Don't accept start with a dash */
    rc = check_senderhelo_char_start(senderhelo, action, '-');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    if (len > 1) {
      /* Don't accept end with a dot */
      rc = check_senderhelo_char_end(senderhelo, action, '.');
      if (rc == 2) return &resp;
      if (rc == 1) return 0;
      /* Don't accept end with a ] */
      rc = check_senderhelo_char_end(senderhelo, action, ']');
      if (rc == 2) return &resp;
      if (rc == 1) return 0;
      /* Don't accept end with a dash */
      rc = check_senderhelo_char_end(senderhelo, action, '-');
      if (rc == 2) return &resp;
      if (rc == 1) return 0;
    }
    if (strchr(senderhelo, '.') == NULL) {
      if (action > 0) {
        resp.message = "Sorry, your HELO name doesn't contains the character '.'";
        return &resp;
      } else {
        elogerr(logcheck, 1, "'.' character not found in HELO name %s for REMOTEIP %s\n", senderhelo, remote_ip);
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: '.' character not found in HELO name ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "\n");
        return 0;
      }
    }
    rc = check_senderhelo_char(senderhelo, action, '!');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '@');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '#');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '$');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '%');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '^');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '&');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '*');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '(');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ')');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '[');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ']');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '{');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '}');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ':');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ';');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '"');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '\'');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '<');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '>');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '?');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '/');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '|');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '\\');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '+');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '=');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '_');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ' ');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '\t');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ',');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    int okipv4 = is_valid_ipv4(senderhelo);
    if (okipv4 > 0) {
      if (action > 0) {
        resp.message = "Sorry, your HELO name is a valid IPv4 address which is disallowed";
        return &resp;
      } else {
        elogerr(logcheck, 1, "valid IPv4 address found in HELO name %s for REMOTEIP %s\n", senderhelo, remote_ip);
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: Valid IPv4 address found in HELO name ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "\n");
        return 0;
      }
    }
  }
  int matchipcount = 0;
  int matchip3count = 0;
  char *matchip = NULL;
  char *senderheloip = NULL;
  struct HostDNSIP hh;
  init_HostDNSIP(&hh);
  elogerr(logcheck, 1, "Start doing DNS Lookup for HELO name %s REMOTEIP %s\n", senderhelo, remote_ip);
  int ret = check_hostname(senderhelo, &hh, MAX_DNSIP_RECORDS);
  if (ret > 0) {
    int x = 0;
    for (x = 0; x < hh.ipcount && hh.ip[x] != NULL; x++) {
      elogerr(logcheck, 1, "ip[%d]=%s ptr[%d]=%s", x, hh.ip[x], x, hh.ptr[x]);
      if (!senderheloip) {
        senderheloip = hh.ip[x];
      }
      /* If it is different then it is a ptr */
      if (strcmp(hh.ip[x], hh.ptr[x]) != 0) {
        elogerr(logcheck, 0, " ptr=YES");
      } else {
        elogerr(logcheck, 0, " ptr=NO");
      }
      /* matching first 3 octets of IPv4 */
      if (!strchr(hh.ip[x], ':')) {
        if (strcmp(remote_ip, hh.ip[x]) == 0) {
          elogerr(logcheck, 0, " ipmatch=YES");
          matchipcount++;
        }
        else if (strcmp(remote_ip, hh.ip[x])) {
          if (match_ipv4(remote_ip, hh.ip[x], 3)) {
            elogerr(logcheck, 0, " ipmatch3=YES");
            if (!matchip) {
              matchip = hh.ip[x];
            }
            matchip3count++;
          } else {
            elogerr(logcheck, 0, " ipmatch3=NO");
          }
        } else {
          elogerr(logcheck, 0, " ipmatch=NO");
        }
      }
      /* If it is the same then it is valid ptr */
      if (strcmp(hh.hostname, hh.ptr[x]) == 0) {
        elogerr(logcheck, 0, " ptrcheck=PASSED\n");
      } else {
        elogerr(logcheck, 0, " ptrcheck=FAILED\n");
      }
    }
    if (checkremoteip || checkremoteipmatch) {
      if (checkremoteip && checkremoteipmatch) {
        elogerr(logcheck, 1, "$HELO_CHECK_REMOTEIP and $HELO_CHECK_REMOTE IP_MATCH are set for REMOTEIP %s\n", remote_ip);
      } else if (checkremoteip) {
        elogerr(logcheck, 1, "$HELO_CHECK_REMOTEIP is set for REMOTEIP %s\n", remote_ip);
      } else {
        elogerr(logcheck, 1, "$HELO_CHECK_REMOTEIP_MATCH is set for REMOTEIP %s\n", remote_ip);
      }
      if (matchipcount > 0) {
        elogerr(logcheck, 1, "REMOTEIP and HELO name %s's IP %s match\n", senderhelo, remote_ip);
        str_cats(&hmsg, "PASSED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEIP ");
        str_cats(&hmsg, remote_ip);
        str_cats(&hmsg, " matches ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "'s IP\n");
        (void)free_HostDNSIP(&hh);
        return 0;
      } else if (matchip3count > 0 && checkremoteipmatch && matchip) {
        elogerr(logcheck, 1, "REMOTEIP %s match the first 3 octets of HELO name %s's IP %s (same class C)\n", remote_ip, senderhelo, matchip);
        str_cats(&hmsg, "SKIPPED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEIP ");
        str_cats(&hmsg, remote_ip);
        str_cats(&hmsg, " matches first ");
        str_cats(&hmsg, "3 octets of ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "'s IP ");
        str_cats(&hmsg, matchip);
        str_cats(&hmsg, " (same class C)\n");
        (void)free_HostDNSIP(&hh);
        return 0;
      }
      if (checkremoteipmatch) {
        elogerr(logcheck, 1, "REMOTEIP %s doesn't match the first 3 octets of HELO name %s's IP %s (same class C)\n", remote_ip, senderhelo, senderheloip);
        if (action > 0) {
          str_init(&estr);
          str_cats(&estr, "Sorry, your REMOTEIP ");
          str_cats(&estr, remote_ip);
          str_cats(&estr, " doesn't match your HELO name's IP or within the same class C");
          str_cats(&estr, senderheloip);
          resp.message = estr.s;
          return &resp;
        } else {
          str_cats(&hmsg, "FAILED\n");
          str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEIP ");
          str_cats(&hmsg, remote_ip);
          str_cats(&hmsg, " doesn't match the first 3 octets of ");
          str_cats(&hmsg, senderhelo);
          str_cats(&hmsg, "'s IP ");
          str_cats(&hmsg, senderheloip);
          str_cats(&hmsg, " (same class C)\n");
          return 0;
        }
      }
      elogerr(logcheck, 1, "REMOTEIP %s doesn't match HELO name %s's IP %s\n", remote_ip, senderhelo, senderheloip);
      if (action > 0) {
        str_init(&estr);
        str_cats(&estr, "Sorry, your REMOTEIP ");
        str_cats(&estr, remote_ip);
        str_cats(&estr, " doesn't match your HELO name's IP ");
        str_cats(&estr, senderheloip);
        resp.message = estr.s;
        (void)free_HostDNSIP(&hh);
        return &resp;
      } else {
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEIP ");
        str_cats(&hmsg, remote_ip);
        str_cats(&hmsg, " doesn't match ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "'s IP ");
        str_cats(&hmsg, senderheloip);
        str_cats(&hmsg, "\n");
        (void)free_HostDNSIP(&hh);
        return 0;
      }
    }
    const char *helo_ptrcheck = session_getenv("HELO_CHECK_PTR");
    if (hh.ptrcount == 0 && helo_ptrcheck) {
      if (action > 0) {
        resp.message = "Sorry, your HELO name doesn't has a valid PTR record";
        (void)free_HostDNSIP(&hh);
        return &resp;
      } else {
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: PTR record not found for HELO name ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "\n");
        (void)free_HostDNSIP(&hh);
        return 0;
      }
    }
    const char *helo_ptrcheck_match = session_getenv("HELO_CHECK_PTR_MATCH");
    if (hh.ptrcheck == 0 && helo_ptrcheck_match) {
      if (action > 0) {
        resp.message = "Sorry, your HELO name doesn't has a valid PTR record that resolves back to the same name";
        (void)free_HostDNSIP(&hh);
        return &resp;
      } else {
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: PTR record doesn't resolves back to HELO name ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "\n");
        (void)free_HostDNSIP(&hh);
        return 0;
      }
    }
    if (hh.ptrcheck > 0) {
      str_cats(&hmsg, "PASSED\n");
    } else {
      str_cats(&hmsg, "FAILED\n");
    }
    char tt[512];
    str_cats(&hmsg, "X-HELO-Check-Summary: ");
    str_cats(&hmsg, hh.hostname);
    str_cats(&hmsg, " has ");
    sprintf(tt, "%d", hh.ipv4count);
    str_cats(&hmsg, tt);
    str_cats(&hmsg, " A ");
    if (hh.ipv4count > 1) {
      str_cats(&hmsg, "records");
    } else {
      str_cats(&hmsg, "record");
    }
    str_cats(&hmsg, ", ");
    sprintf(tt, "%d", hh.ipv6count);
    str_cats(&hmsg, tt);
    str_cats(&hmsg, " AAAA ");
    if (hh.ipv6count > 1) {
      str_cats(&hmsg, "records");
    } else {
      str_cats(&hmsg, "record");
    }
    str_cats(&hmsg, ", ");
    sprintf(tt, "%d", hh.ptrcount);
    str_cats(&hmsg, tt);
    str_cats(&hmsg, " PTR ");
    if (hh.ptrcount > 1) {
      str_cats(&hmsg, "records");
    } else {
      str_cats(&hmsg, "record");
    }
    str_cats(&hmsg, " and ");
    sprintf(tt, "%d", hh.ptrcheck);
    str_cats(&hmsg, tt);
    str_cats(&hmsg, " PTR ");
    if (hh.ptrcheck > 1) {
      str_cats(&hmsg, "records");
    } else {
      str_cats(&hmsg, "record");
    }
    str_cats(&hmsg, " resolves to the same hostname\n");
    (void)free_HostDNSIP(&hh);
    return 0;
  }
  if (outofmemory > 0 || ret == EAI_MEMORY) {
    eprintf_logname();
    eprintf("(%s:%d) Out of Memory!\n", __FILE__, __LINE__);
    return &resp_oom;
  }
  char *tmpmsg = (char *)malloc(512);
  if (tmpmsg == NULL) {
    elogerr(1, 1, "(%s:%d) char *tmpmsg is out of memory!\n", __FILE__, __LINE__);
    return &resp_oom;
  }
  if (action > 0) {
    if (ret == EAI_NONAME) {
      sprintf(tmpmsg, "Sorry, your HELO name is invalid.");
      if (strlen(senderhelo) > 0) {
        sprintf(tmpmsg, "Sorry, your HELO name %s is invalid.", senderhelo);
      }
      str_init(&estr);
      str_cats(&estr, tmpmsg);
      resp.message = estr.s;
      free(tmpmsg);
      return &resp;
    } else {
      sprintf(tmpmsg, "Sorry, your HELO name DNS lookup returned a temporary failure. Please try again later.");
      if (strlen(senderhelo) > 0) {
        sprintf(tmpmsg, "Sorry, your HELO name %s DNS lookup returned a temporary failure. Please try again later.", senderhelo);
      }
      resp.number = 451;
      str_init(&estr);
      str_cats(&estr, tmpmsg);
      resp.message = estr.s;
      free(tmpmsg);
      return &resp;
    }
  } else {
    str_cats(&hmsg, "FAILED\n");
    elogerr(logcheck, 1, "(%s:%d) HELO name %s DNS lookup error\n", __FILE__, __LINE__, senderhelo);
    str_cats(&hmsg, "X-HELO-Check-Summary: DNS lookup error for HELO name ");
    if (strlen(senderhelo) > 0) {
      str_cats(&hmsg, senderhelo);
    }
    str_cats(&hmsg, "\n");
    free(tmpmsg);
    return 0;
  }
  return 0;
  (void)address;
  (void)params;
}

static const response *data_start(int fd)
{
  if (skipcheck) return 0;
  if (hmsg.len > 0 && addheadercheck) {
    return backend_data_block(hmsg.s, hmsg.len);
  }
  str_init(&hmsg);
  return 0;
  (void)fd;
}

struct plugin plugin = {
  .version = PLUGIN_VERSION,
  .flags = 0,
  .init = init,
  .helo = helo,
  .sender = sender,
  .data_start = data_start,
};
